import themes
import imgmap
import txtmap
import stretch
import os
import time
from pathlib import Path


# ==================== List of menu option functions ==================== #
def setTSV(args):
    print("Current TSV file:", args.MAPFILE)
    print("Enter in the location + name of your TSV file:")
    args.MAPFILE = str(Path(input()).expanduser())
    return args


def setTheme(args):
    print("Current theme:", args.tileset)
    print("Get inbuilt theme? (y/N)")

    getNative = input()
    if getNative == 'y' or getNative == 'Y':
        themes.printThemes()
        tmpTheme = themes.selTheme(args.tileset)
        if tmpTheme is not False:
            args.tileset = tmpTheme
    else:
        print("Enter in your theme directory:")
        args.tileset = str(Path(input()).expanduser())

    return args


def setSave(args):
    print("Current save directory:", args.savetiles)
    print("Enter [name]/[location]/[location + name] to save your map image:")
    args.output = input()
    return args


def setSize(args):
    print("Current tile size:", args.pixels)
    print("Enter in the new size of your tiles:")

    newSize = input()
    if newSize.isdigit():
        args.pixels = int(newSize)
        print("Size set to", newSize)
    else:
        print("Invalid input! (Press Enter to continue.)")
        input()

    return args


def togMetric(args):
    args.measure ^= True
    return args


def togRandom(args):
    args.randomise ^= True
    return args


def genTheme(args):
    if args.tileset is not None:
        print("Attempting theme generation at", args.tileset)
        start = time.time()

        tmpRes = themes.canGenerate(args.tileset, True)
        tmpRes = themes.generateTheme(tmpRes)
        if themes.writeTheme(args.tileset, tmpRes) is False:
            print("Insufficient resources found. (Press Enter to continue)")
        else:
            if args.measure:
                end = time.time()
                print("Done in", end - start, "seconds.")
            print("Complete! (Press Enter to continue)")
            input()
    else:
        print("Specify tileset folder! (Press Enter to continue)")
        input()
    return args


def genMap(args):
    if args.MAPFILE is not None and args.tileset is not None:
        print("Attempting map generation to", args.output)
        start = time.time()

        tmpTheme = themes.canGenerate(args.tileset, True)
        tmpTheme = themes.generateTheme(tmpTheme)

        tmpMap = txtmap.readMap(args.MAPFILE)
        tmpMap = txtmap.parseMap(tmpMap)
        tmpMap = imgmap.generateMap(args, tmpTheme, tmpMap)

        if imgmap.writeMap(args, tmpMap) is False:
            print("Insufficient resources found. (Press Enter to continue)")
            input()
        else:
            if args.measure:
                end = time.time()
                print("Done in", end - start, "seconds.")
            print("Complete! (Press Enter to continue)")
            input()
    else:
        print("Specify more details! (Press Enter to continue)")
        input()
    return args


def progExit():
    pass


choices = {
        "Select TSV File": setTSV,
        "Select Theme": setTheme,
        "Set Save Location": setSave,
        "Set Tile Size": setSize,
        "Toggle Savetime Metrics": togMetric,
        "Toggle Floor Shuffling": togRandom,
        "Generate Theme": genTheme,
        "Generate Map Image": genMap,
        "Exit Generator": progExit
    }
# ==================== End of map menu options ==================== #


def getSettings(args):
    optList = {
        "TSV Map File\t\t:": args.MAPFILE,
        "Tile Theme Folder\t:": args.tileset,
        "Save Location\t\t:": args.output,
        "Tile Size (Pixels)\t:": args.pixels,
        "Measure Save Time\t:": args.measure,
        "Shuffle Floor\t\t:": args.randomise
    }
    for i, val in optList.items():
        print(i, val)


def getOptions():
    for i, (key, val) in enumerate(choices.items()):
        print('[' + str(i) + ']', key)


def mainmenu(args):
    option = None
    while option != (len(choices) - 1):
        os.system('cls' if os.name == 'nt' else 'clear')
        print("========== Current Settings ==========")
        getSettings(args)
        print()
        getOptions()
        print("Select option: ")

        tmpOpt = input()
        # Execute linked option
        if tmpOpt.isdigit():
            optList = list(choices.keys())
            option = int(tmpOpt)

            # Parse arguments
            args = stretch.expandargs(args)

            # No arguments for exiting
            if option < (len(optList) - 1):
                args = choices[optList[option]](args)
            elif option == (len(optList) - 1):
                choices[optList[option]]
    print("Exiting...")



from pathlib import Path
from PIL import Image
from random import choice


# Actually create the map in question
def generateMap(args, mapBlocks, resFile):
    if mapBlocks is not False and resFile != []:
        # Scale theme tiles to desired size
        for supKey, category in mapBlocks.items():
            for key, img in category.items():
                mapBlocks[supKey][key] = mapBlocks[supKey][key].resize(
                    (args.pixels, args.pixels),
                    resample=Image.LANCZOS)

        # Convert all tiles to RGBA mode
        for supKey, category in mapBlocks.items():
            for key, img in category.items():
                if mapBlocks[supKey][key].mode != 'RGBA':
                    mapBlocks[supKey][key] = mapBlocks[supKey][key].convert(
                        'RGBA'
                    )

        # Get width & height of image
        width = len(resFile[0]) * args.pixels
        height = len(resFile) * args.pixels

        # Paste images onto blank canvas
        tmpImage = Image.new('RGBA', (width, height))
        if mapBlocks is not False:
            for row, tileMap in enumerate(resFile):
                yPos = row * args.pixels
                # Loop through tile resources in single tile
                for col, tileArray in enumerate(tileMap):
                    xPos = col * args.pixels
                    # Merge resources if there's more than one
                    if len(tileArray) > 1:
                        tileImage = Image.new(
                            'RGBA', (args.pixels, args.pixels)
                        )
                        for pos, imgCur in enumerate(tileArray):
                            if pos == (len(tileArray) - 1):
                                break
                            imgNext = tileArray[pos+1]
                            bg = mapBlocks[imgCur[0]][imgCur[1]]
                            fg = mapBlocks[imgNext[0]][imgNext[1]]
                            tileImage = Image.alpha_composite(
                                tileImage, bg)
                            tileImage = Image.alpha_composite(
                                tileImage, fg)
                        tmpImage.paste(im=tileImage, box=(xPos, yPos))
                    else:
                        # Floor tile randomising
                        tileCat = tileArray[0][0]
                        tileNam = tileArray[0][1]
                        if args.randomise and tileNam == 'floor':
                            floorImg = mapBlocks[tileCat][tileNam]
                            floorImg = choice([
                                floorImg,
                                floorImg.transpose(Image.ROTATE_90),
                                floorImg.transpose(Image.ROTATE_180),
                                floorImg.transpose(Image.ROTATE_270)
                            ])
                            tmpImage.paste(im=floorImg, box=(xPos, yPos))
                        else:
                            tmpImage.paste(
                                im=mapBlocks[tileCat][tileNam],
                                box=(xPos, yPos)
                            )
        return tmpImage
    else:
        return False


# Save map to a specific directory
def writeMap(args, tmpMap):

    if tmpMap is not False:
        defName = 'Map.png'
        defLocs = Path.home()
        saveLoc = Path(args.output).expanduser()

        fileTypes = ['.png', '.jpg', '.jpeg']

        if saveLoc.is_dir():
            saveLoc = Path(saveLoc, defName)
            tmpMap.save(saveLoc)
        elif saveLoc.parent.is_dir():
            if saveLoc.suffix in fileTypes:
                tmpMap.save(saveLoc)
            else:
                saveLoc = str(saveLoc) + '.png'
                tmpMap.save(saveLoc)
        else:
            saveLoc = str(Path(defLocs, defName))
            tmpMap.save(saveLoc)
    else:
        return False



from pathlib import Path
from PIL import Image
import copy


# Retrieve list of theme directories
def getThemes():
    themeList = []

    imgLoc = Path(__file__).parent.resolve()
    imgLoc = Path(imgLoc, 'themes')

    for item in imgLoc.iterdir():
            if item.is_dir():
                themeList.append(item)

    themeList = sorted(themeList)
    return themeList


# Display all builtin themes
def printThemes():
    curList = getThemes()
    print("===Theme List===")
    for key, item in enumerate(curList):
        themeName = item.parts
        print("[" + str(key) + "]: " + themeName[-1])


# Return desired theme folder
def selTheme(curTile):
    themes = getThemes()
    options = range(0, len(themes))
    themeIndex = list(str(item) for item in options)
    selection = input()

    if selection in themeIndex:
        curTile = themes[int(selection)]
    else:
        curTile = False
    return curTile


# ==================== Theme Generation Functions ====================
resources = {
    'floorAssets': {
        'floor':                '0-Floor'
    },
    'spaceAssets': {
        'dungeonSpace':         '1-Space'
    },
    'wallAssets': {
        'wallTop':              '2-T_Wall',
        'wallBottom':           '2-B_Wall',
        'wallLeft':             '2-L_Wall',
        'wallRight':            '2-R_Wall'
    },
    'inCorners': {
        'topLeftCornerI':       '3-TL_Corner_I',
        'bottomRightCornerI':   '3-BR_Corner_I',
        'bottomLeftCornerI':    '3-BL_Corner_I',
        'topRightCornerI':      '3-TR_Corner_I'
    },
    'outCorners': {
        'topLeftCornerO':       '4-TL_Corner_O',
        'bottomRightCornerO':   '4-BR_Corner_O',
        'bottomLeftCornerO':    '4-BL_Corner_O',
        'topRightCornerO':      '4-TR_Corner_O'
    },
    'doorAssets': {
        'doorTop':              '5-T_Door',
        'doorBottom':           '5-B_Door',
        'doorLeft':             '5-L_Door',
        'doorRight':            '5-R_Door'
    },
    'doorSAssets': {
        'doorSecretTop':        '6-T_Door_Secret',
        'doorSecretBottom':     '6-B_Door_Secret',
        'doorSecretLeft':       '6-L_Door_Secret',
        'doorSecretRight':      '6-R_Door_Secret'
    },
    'doorPAssets': {
        'doorPortTop':          '7-T_Door_Portcullis',
        'doorPortBottom':       '7-B_Door_Portcullis',
        'doorPortLeft':         '7-L_Door_Portcullis',
        'doorPortRight':        '7-R_Door_Portcullis'
    },
    'stairUAssets': {
        'topStairU':            '8-T_Stair_U',
        'bottomStairU':         '8-B_Stair_U',
        'leftStairU':           '8-L_Stair_U',
        'rightStairU':          '8-R_Stair_U'
    },
    'stairUUAssets': {
        'topStairUU':           '9-T_Stair_UU',
        'bottomStairUU':        '9-B_Stair_UU',
        'leftStairUU':          '9-L_Stair_UU',
        'rightStairUU':         '9-R_Stair_UU'
    },
    'stairDAssets': {
        'topStairD':            '10-T_Stair_D',
        'bottomStairD':         '10-B_Stair_D',
        'leftStairD':           '10-L_Stair_D',
        'rightStairD':          '10-R_Stair_D'
    },
    'stairDDAssets': {
        'topStairDD':           '11-T_Stair_DD',
        'bottomStairDD':        '11-B_Stair_DD',
        'leftStairDD':          '11-L_Stair_DD',
        'rightStairDD':         '11-R_Stair_DD'
    }
}


# Check if resource generation is possible
def canGenerate(themePath, openImages=False):
    tmpRes = copy.deepcopy(resources)

    if Path(themePath).is_dir:
        sufficient = []
        fileTypes = ['.png', '.jpg', '.jpeg']
        # Different asset types
        for superKey, val in tmpRes.items():
            # Asset type members
            state = False
            for assetKey, fileName in val.items():
                for suffix in fileTypes:
                    tmpName = fileName + suffix
                    if Path(themePath, tmpName).is_file():
                        if openImages:
                            tmpRes[superKey][assetKey] = Image.open(
                                Path(themePath, tmpName)
                            )
                        state = True
            sufficient.append(state)

        if openImages:
            return [all(sufficient), tmpRes]
        else:
            return all(sufficient)
    else:
        return False


# Supply a fully made theme if possible
def generateTheme(tmpRes):
    if tmpRes[0] is not False:
        # Dictionary with some open images
        tmpRes = tmpRes[1]

        def rotateNone(im):
            return im

        def rotateLeft(im):
            tmpImage = im.transpose(Image.ROTATE_90)
            tmpImage.format = im.format
            return tmpImage

        def rotateFlip(im):
            tmpImage = im.transpose(Image.ROTATE_180)
            tmpImage.format = im.format
            return tmpImage

        def rotateRight(im):
            tmpImage = im.transpose(Image.ROTATE_270)
            tmpImage.format = im.format
            return tmpImage

        rotate = {
            0: {
                0: rotateNone,
                1: rotateFlip,
                2: rotateLeft,
                3: rotateRight
            },
            1: {
                0: rotateFlip,
                1: rotateNone,
                2: rotateRight,
                3: rotateLeft
            },
            2: {
                0: rotateRight,
                1: rotateLeft,
                2: rotateNone,
                3: rotateFlip
            },
            3: {
                0: rotateLeft,
                1: rotateRight,
                2: rotateFlip,
                3: rotateNone
            }
        }

        # Different asset types -> asset permutations
        for key, val in tmpRes.items():
            # Get asset to transpose to others
            imgIndex = ''
            transposee = ''
            for i, (assetKey, img) in enumerate(val.items()):
                if isinstance(img, Image.Image):
                    imgIndex = i
                    transposee = img
                    break
            # Create other assets from found one
            for i, (assetKey, img) in enumerate(val.items()):
                if not isinstance(img, Image.Image):
                    tmpRes[key][assetKey] = rotate[imgIndex][i](transposee)

        return tmpRes
    else:
        return False


# Create a theme at the specified path
def writeTheme(themePath, themeRes):
    if themeRes is not False:
        for itemGroup, category in themeRes.items():
            for item, asset in category.items():
                imgName = resources[itemGroup][item]
                imgName = imgName + '.' + asset.format.lower()
                saveLoc = Path(themePath, imgName)
                if not saveLoc.is_file():
                    asset.save(saveLoc)
        print("Theme saved!")
    else:
        return False
